@@ -15,6 +15,8 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <assert.h>
+#include <glib.h>
 #include <libguile.h>
 
 #include "../app.h"
@@ -40,6 +42,18 @@ gzochid_application_context *gzochid_api_ensure_current_application_context ()
   return context;
 }
 
+void gzochid_api_check_not_found (GError *err)
+{
+  assert (err != NULL);
+  if (err->code == GZOCHID_DATA_ERROR_NOT_FOUND)
+    {
+      g_error_free (err);
+      gzochid_scheme_r6rs_raise_continuable
+	(gzochid_scheme_make_object_removed_condition ());
+    }
+  else g_error_free (err);
+}
+
 void gzochid_api_check_transaction ()
 {
   if (gzochid_transaction_timed_out ())
@@ -49,8 +63,15 @@ void gzochid_api_check_transaction ()
 		   scm_call_0 (scm_make_transaction_retry_condition)));
 
   if (gzochid_transaction_rollback_only ())
-    gzochid_scheme_r6rs_raise 
-      (scm_call_0 (scm_make_transaction_aborted_condition));
+    {
+      if (gzochid_transaction_retryable ())
+	gzochid_scheme_r6rs_raise 
+	  (scm_call_2 (scm_condition, 
+		       scm_call_0 (scm_make_transaction_aborted_condition),
+		       scm_call_0 (scm_make_transaction_retry_condition)));
+      else gzochid_scheme_r6rs_raise 
+	     (scm_call_0 (scm_make_transaction_aborted_condition));
+    }  
 }
 
 static void bind_scm (char *module, SCM *binding, char *name)
