@@ -15,7 +15,10 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+#include <assert.h>
 #include <glib.h>
+#include <gmp.h>
+#include <stddef.h>
 #include <stdlib.h>
 #include <sys/time.h>
 
@@ -118,6 +121,8 @@ static void initialize_data (int from_state, int to_state, gpointer user_data)
   char *oids_db = g_strconcat (data_dir, "/oids", NULL);
   char *names_db = g_strconcat (data_dir, "/names", NULL);
 
+  gzochid_storage_context *storage_context = NULL;
+
   if (!g_file_test (data_dir, G_FILE_TEST_EXISTS))
     {
       gzochid_notice 
@@ -128,9 +133,12 @@ static void initialize_data (int from_state, int to_state, gpointer user_data)
   else if (!g_file_test (data_dir, G_FILE_TEST_IS_DIR))
     gzochid_err ("%s is not a directory.", data_dir);
 
-  app_context->meta = gzochid_storage_open (meta_db);
-  app_context->oids = gzochid_storage_open (oids_db);
-  app_context->names = gzochid_storage_open (names_db);
+  storage_context = gzochid_storage_initialize (data_dir);
+
+  app_context->storage_context = storage_context;
+  app_context->meta = gzochid_storage_open (storage_context, meta_db);
+  app_context->oids = gzochid_storage_open (storage_context, oids_db);
+  app_context->names = gzochid_storage_open (storage_context, names_db);
 
   free (data_dir);
   free (meta_db);
@@ -145,7 +153,7 @@ static void initialize_load_paths_guile_worker
 
   while (load_path_ptr != NULL)
     {
-      gzochid_scheme_append_load_path ((char *) load_path_ptr->data);
+      gzochid_guile_add_to_load_path ((char *) load_path_ptr->data);
       load_path_ptr = load_path_ptr->next;
     }
 }
@@ -316,12 +324,14 @@ gzochid_transaction_result gzochid_application_transaction_execute_timed
 }
 
 static gzochid_transactional_application_task_execution *execution_new
-(gzochid_application_task *task, struct timeval *timeout)
+(gzochid_application_task *task, gzochid_application_task *cleanup_task,
+ struct timeval *timeout)
 {
   gzochid_transactional_application_task_execution *execution = 
     malloc (sizeof (gzochid_transactional_application_task_execution));
 
   execution->task = task;
+  execution->cleanup_task = cleanup_task;
 
   if (timeout != NULL)
     {
@@ -339,16 +349,17 @@ static gzochid_transactional_application_task_execution *execution_new
 
 gzochid_transactional_application_task_execution *
 gzochid_transactional_application_task_execution_new 
-(gzochid_application_task *task)
+(gzochid_application_task *task, gzochid_application_task *cleanup_task)
 {
-  return execution_new (task, NULL);
+  return execution_new (task, NULL, NULL);
 }
 
 gzochid_transactional_application_task_execution *
 gzochid_transactional_application_task_timed_execution_new 
-(gzochid_application_task *task, struct timeval timeout)
+(gzochid_application_task *task, gzochid_application_task *cleanup_task,
+ struct timeval timeout)
 {
-  return execution_new (task, &timeout);
+  return execution_new (task, cleanup_task, &timeout);
 }
 
 void gzochid_transactional_application_task_execution_free
@@ -402,30 +413,44 @@ void gzochid_application_resubmitting_transactional_task_worker
   gzochid_application_transactional_task_worker 
     (app_context, identity, execution);
   
-  if (execution->result != GZOCHID_TRANSACTION_SUCCESS
-      && gzochid_application_should_retry (execution))
+  if (execution->result != GZOCHID_TRANSACTION_SUCCESS)
     {
       struct timeval now;
       gzochid_context *context = (gzochid_context *) app_context;
       gzochid_game_context *game_context = 
 	(gzochid_game_context *) context->parent;
-
       gzochid_application_task *application_task = NULL;
       gzochid_task *task = NULL;
 
-      execution->result = GZOCHID_TRANSACTION_PENDING;
+      if (gzochid_application_should_retry (execution))
+	{	  	  
+	  execution->result = GZOCHID_TRANSACTION_PENDING;
+	  
+	  application_task = gzochid_application_task_new
+	    (app_context, identity,
+	     gzochid_application_resubmitting_transactional_task_worker, 
+	     execution);	  
+	}
+      else if (execution->cleanup_task != NULL)
+	{
+	  gzochid_transactional_application_task_execution *cleanup_execution = 
+	    gzochid_transactional_application_task_execution_new
+	    (execution->cleanup_task, NULL);
 
-      application_task = gzochid_application_task_new
-	(app_context, identity,
-	 gzochid_application_resubmitting_transactional_task_worker, 
-	 execution);
+	  application_task = gzochid_application_task_new
+	    (app_context, identity,
+	     gzochid_application_transactional_task_worker, cleanup_execution);
+	}
 
-      gettimeofday (&now, NULL);
-
-      task = gzochid_task_new
-	(gzochid_application_task_thread_worker, application_task, now);
-
-      gzochid_schedule_submit_task (game_context->task_queue, task);
+      if (application_task != NULL)
+	{
+	  gettimeofday (&now, NULL);
+	  
+	  task = gzochid_task_new
+	    (gzochid_application_task_thread_worker, application_task, now);
+	  
+	  gzochid_schedule_submit_task (game_context->task_queue, task);
+	}
     }
 }
 
@@ -449,15 +474,18 @@ gboolean gzochid_application_should_retry
 
 static void run_async_transactional (gpointer data)
 {
+  GError *err = NULL;
   gzochid_application_context *context = (gzochid_application_context *) data;
-  SCM persisted_callback = (SCM) gzochid_data_get_binding 
-    (context, "s.initializer", &gzochid_scheme_data_serialization);
+  gboolean initialized = gzochid_data_binding_exists 
+    (context, "s.initializer", &err);
 
   gzochid_auth_identity *system_identity = calloc 
     (1, sizeof (gzochid_auth_identity));
   system_identity->name = "[SYSTEM]";
 
-  if (persisted_callback == NULL)
+  assert (err == NULL);
+
+  if (!initialized)
     {
       gzochid_application_task application_task;
       gzochid_task task;
@@ -673,7 +701,7 @@ void gzochid_application_client_logged_in
   gzochid_application_task application_task;
   gzochid_transactional_application_task_execution *execution =
     gzochid_transactional_application_task_timed_execution_new 
-    (&transactional_task, game_context->tx_timeout);
+    (&transactional_task, NULL, game_context->tx_timeout);
 
   gzochid_task task;
 
@@ -720,6 +748,10 @@ void gzochid_application_client_logged_in
 	      g_mutex_unlock (context->client_mapping_lock);
 
 	      gzochid_protocol_client_disconnect (client);
+	      
+	      transactional_task.worker = 
+		gzochid_client_session_disconnected_worker;
+	      gzochid_schedule_run_task (game_context->task_queue, &task);      
 	    }
 	    break;
 	}
@@ -733,7 +765,6 @@ void gzochid_application_client_disconnected
 {
   gzochid_game_context *game_context = 
     (gzochid_game_context *) ((gzochid_context *) context)->parent;
-
   char *session_oid_str = NULL;
 
   g_mutex_lock (context->client_mapping_lock);
@@ -745,31 +776,26 @@ void gzochid_application_client_disconnected
     }
   else 
     {
-      gzochid_application_task *transactional_task =
-	gzochid_application_task_new
+      gzochid_task *callback_task = NULL;
+      gzochid_task *cleanup_task = NULL;
+
+      struct timeval n;
+
+      gettimeofday (&n, NULL);
+      
+      callback_task = gzochid_task_make_transactional_application_task
 	(context, client->identity, 
-	 gzochid_scheme_application_disconnected_worker, session_oid_str);
-      gzochid_transactional_application_task_execution *execution = 
-	gzochid_transactional_application_task_timed_execution_new
-	(transactional_task, game_context->tx_timeout);
-      gzochid_application_task *application_task = gzochid_application_task_new
-	(context, client->identity,
-	 gzochid_application_resubmitting_transactional_task_worker, 
-	 execution);
-
-      gzochid_task *task = NULL;
-      struct timeval now;
-
-      gettimeofday (&now, NULL);
+	 gzochid_scheme_application_disconnected_worker, session_oid_str, n);
+      cleanup_task = gzochid_task_make_transactional_application_task
+	(context, client->identity, gzochid_client_session_disconnected_worker,
+	 session_oid_str, n);
 
       g_hash_table_remove (context->clients_to_oids, client);
       g_hash_table_remove (context->oids_to_clients, session_oid_str);
       
-      task = gzochid_task_new
-	(gzochid_application_task_thread_worker, application_task,
-	 now);
-      
-      gzochid_schedule_submit_task (game_context->task_queue, task);
+      gzochid_schedule_submit_task (game_context->task_queue, callback_task);
+      gzochid_schedule_submit_task (game_context->task_queue, cleanup_task);
+
       g_mutex_unlock (context->client_mapping_lock);
     }
 }
@@ -810,7 +836,7 @@ void gzochid_application_session_received_message
       gzochid_application_task application_task;
       gzochid_transactional_application_task_execution *execution =
 	gzochid_transactional_application_task_timed_execution_new
-	(&transactional_task, game_context->tx_timeout);
+	(&transactional_task, NULL, game_context->tx_timeout);
       gzochid_task task;
 
       data[0] = session_oid_str;
