@@ -18,6 +18,7 @@
 #include <assert.h>
 #include <glib.h>
 #include <gmp.h>
+#include <stddef.h>
 #include <stdlib.h>
 #include <sys/time.h>
 
@@ -210,6 +211,7 @@ static void close_channel
 (gzochid_application_context *context, gzochid_auth_identity *identity, 
  gpointer data)
 {
+  GError *err = NULL;
   gzochid_channel_pending_operation *op = 
     (gzochid_channel_pending_operation *) data;
   gzochid_data_managed_reference *channel_reference =
@@ -217,14 +219,23 @@ static void close_channel
     (context, &gzochid_channel_serialization, op->target_channel);
   gzochid_channel_side_effects_transaction_context *tx_context = NULL;
 
-  char *channel_oid_str = mpz_get_str (NULL, 16, op->target_channel);
+  char *channel_oid_str = NULL; 
   gzochid_channel *channel = NULL;
   GSequenceIter *iter = NULL;
   mpz_t session_oid;
 
-  mpz_init (session_oid);
-  gzochid_data_dereference (channel_reference);
+  gzochid_data_dereference (channel_reference, &err);
+  
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+
   channel = (gzochid_channel *) channel_reference->obj;
+
+  mpz_init (session_oid);
+  channel_oid_str = mpz_get_str (NULL, 16, op->target_channel);
   iter = g_sequence_get_begin_iter (channel->sessions);
   tx_context = join_side_effects_transaction (context, channel);
 
@@ -241,12 +252,31 @@ static void close_channel
 
       session_reference = gzochid_data_create_reference_to_oid 
 	(context, &gzochid_client_session_serialization, session_oid);
-      gzochid_data_dereference (session_reference);
-      session = (gzochid_client_session *) session_reference->obj;
 
+      gzochid_data_dereference (session_reference, &err);
+
+      if (err != NULL)
+	{
+	  if (err->code != GZOCHID_DATA_ERROR_NOT_FOUND)
+	    {
+	      g_error_free (err);
+	      return;
+	    }
+	  else 
+	    {
+	      g_error_free (err);
+	      err = NULL;
+
+	      iter = g_sequence_iter_next (iter);
+	      continue;
+	    }
+	}
+	
+      session = (gzochid_client_session *) session_reference->obj;
       session_iter = g_sequence_lookup
 	(session->channels, channel_oid_str, gzochid_util_string_data_compare,
 	 NULL);
+
       if (session_iter != NULL)
 	{
 	  gzochid_protocol_client *client = g_hash_table_lookup
@@ -261,12 +291,19 @@ static void close_channel
 	  else
 	    {
 	      gzochid_data_mark 
-		(context, &gzochid_channel_serialization, session);
-
-	      tx_context->messages = g_list_append
-		(tx_context->messages,
-		 gzochid_channel_message_new
-		 (GZOCHID_CHANNEL_MESSAGE_LEAVE, client));
+		(context, &gzochid_client_session_serialization, session, 
+		 &err);
+	      
+	      if (err == NULL)
+		tx_context->messages = g_list_append
+		  (tx_context->messages,
+		   gzochid_channel_message_new
+		   (GZOCHID_CHANNEL_MESSAGE_LEAVE, client));
+	      else
+		{
+		  g_error_free (err);
+		  break;
+		}
 	    }
 	}
       
@@ -282,6 +319,7 @@ static void send_channel_message
 (gzochid_application_context *context, gzochid_auth_identity *identity,
  gpointer data)
 {
+  GError *err = NULL;
   gzochid_channel_pending_send_operation *send_op =
     (gzochid_channel_pending_send_operation *) data;
   gzochid_channel_pending_operation *op = 
@@ -295,7 +333,14 @@ static void send_channel_message
   GSequenceIter *iter = NULL;
   gboolean channel_modified = FALSE;
 
-  gzochid_data_dereference (channel_reference);
+  gzochid_data_dereference (channel_reference, &err);
+
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+  
   channel = (gzochid_channel *) channel_reference->obj;
   tx_context = join_side_effects_transaction (context, channel);
   iter = g_sequence_get_begin_iter (channel->sessions);
@@ -327,7 +372,8 @@ static void send_channel_message
     }
 
   if (channel_modified)
-    gzochid_data_mark (context, &gzochid_channel_serialization, channel);
+    gzochid_data_mark 
+      (context, &gzochid_channel_serialization, channel, NULL);
   
   g_mutex_unlock (context->client_mapping_lock);
 }
@@ -336,6 +382,7 @@ static void join_channel
 (gzochid_application_context *context, gzochid_auth_identity *identity,
  gpointer data)
 {
+  GError *err = NULL;
   gzochid_channel_pending_membership_operation *member_op =
     (gzochid_channel_pending_membership_operation *) data;
   gzochid_channel_pending_operation *op = 
@@ -352,15 +399,27 @@ static void join_channel
     (context, &gzochid_client_session_serialization, member_op->target_session);
   gzochid_channel_side_effects_transaction_context *tx_context = NULL;
 
-  char *session_oid_str = mpz_get_str (NULL, 16, session_reference->oid);
+  char *session_oid_str = NULL;
   GSequenceIter *iter = NULL;
 
-  gzochid_data_dereference (channel_reference);
-  gzochid_data_dereference (session_reference);
+  gzochid_data_dereference (channel_reference, &err);
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+
+  gzochid_data_dereference (session_reference, &err);
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
 
   channel = (gzochid_channel *) channel_reference->obj;
   session = (gzochid_client_session *) session_reference->obj;
 
+  session_oid_str = mpz_get_str (NULL, 16, session_reference->oid);
   iter = g_sequence_lookup 
     (channel->sessions, session_oid_str, gzochid_util_string_data_compare, 
      NULL);
@@ -384,9 +443,16 @@ static void join_channel
 	    (tx_context->messages, gzochid_channel_message_new 
 	     (GZOCHID_CHANNEL_MESSAGE_JOIN, client));
 
-	  gzochid_data_mark (context, &gzochid_channel_serialization, channel);
 	  gzochid_data_mark 
-	    (context, &gzochid_client_session_serialization, session);
+	    (context, &gzochid_channel_serialization, channel, &err);
+	  if (err != NULL)
+	    {
+	      g_error_free (err);
+	      return;
+	    }
+	  gzochid_data_mark 
+	    (context, &gzochid_client_session_serialization, session, NULL);
+	  return;
 	}
       else gzochid_warning 
 	     ("Client not found for joined channel session '%s'; skipping.", 
@@ -399,6 +465,7 @@ static void leave_channel
 (gzochid_application_context *context, gzochid_auth_identity *identity,
  gpointer data)
 {
+  GError *err = NULL;
   gzochid_channel_pending_membership_operation *member_op =
     (gzochid_channel_pending_membership_operation *) data;
   gzochid_channel_pending_operation *op = 
@@ -415,15 +482,27 @@ static void leave_channel
     (context, &gzochid_client_session_serialization, member_op->target_session);
   gzochid_channel_side_effects_transaction_context *tx_context = NULL;
 
-  char *session_oid_str = mpz_get_str (NULL, 16, session_reference->oid);
+  char *session_oid_str = NULL;
   GSequenceIter *iter = NULL;
 
-  gzochid_data_dereference (channel_reference);
-  gzochid_data_dereference (session_reference);
+  gzochid_data_dereference (channel_reference, &err);
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+ 
+  gzochid_data_dereference (session_reference, &err);
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
 
   channel = (gzochid_channel *) channel_reference->obj;
   session = (gzochid_client_session *) session_reference->obj;
 
+  session_oid_str = mpz_get_str (NULL, 16, session_reference->oid);
   iter = g_sequence_lookup 
     (channel->sessions, session_oid_str, gzochid_util_string_data_compare, 
      NULL);
@@ -450,9 +529,16 @@ static void leave_channel
 
 	  free (channel_oid_str);
 	  
-	  gzochid_data_mark (context, &gzochid_channel_serialization, channel);
 	  gzochid_data_mark 
-	    (context, &gzochid_client_session_serialization, session);
+	    (context, &gzochid_channel_serialization, channel, &err);
+	  if (err != NULL)
+	    {
+	      g_error_free (err);
+	      return;
+	    }
+	  gzochid_data_mark 
+	    (context, &gzochid_client_session_serialization, session, NULL);
+	  return;
 	}
       else gzochid_warning 
 	     ("Client not found for parting channel session '%s'; skipping.", 
@@ -738,6 +824,7 @@ static char *make_channel_binding (char *name)
 gzochid_channel *gzochid_channel_create
 (gzochid_application_context *context, char *name)
 {
+  GError *err = NULL;
   char *binding = make_channel_binding (name);
   gzochid_channel *channel = gzochid_channel_new (name);
   gzochid_data_managed_reference *reference = gzochid_data_create_reference 
@@ -754,10 +841,17 @@ gzochid_channel *gzochid_channel_create
   channel->id = (unsigned char *) mpz_get_str (NULL, 16, channel->oid);
   channel->id_len = strlen ((char *) channel->id);
 
-  gzochid_data_set_binding_to_oid (context, binding, reference->oid);
-  free (binding);
-
-  return channel;
+  gzochid_data_set_binding_to_oid (context, binding, reference->oid, &err);
+  if (err != NULL)
+    {
+      free (binding);
+      return NULL;
+    }
+  else 
+    {
+      free (binding);
+      return channel;
+    }
 }
 
 gzochid_channel *gzochid_channel_get 
@@ -765,7 +859,7 @@ gzochid_channel *gzochid_channel_get
 {
   char *binding = make_channel_binding (name);
   gzochid_channel *channel = (gzochid_channel *) gzochid_data_get_binding
-    (context, binding, &gzochid_channel_serialization);
+    (context, binding, &gzochid_channel_serialization, NULL);
 
   free (binding);
 
