@@ -18,6 +18,7 @@
 #include <assert.h>
 #include <glib.h>
 #include <gmp.h>
+#include <stddef.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -33,6 +34,11 @@
 
 #define NEXT_OID_KEY 0x00
 
+GQuark gzochid_data_error_quark (void)
+{
+  return g_quark_from_static_string ("gzochid-data-error-quark");
+}
+
 gzochid_oid_holder *gzochid_oid_holder_new (void)
 {
   gzochid_oid_holder *holder = malloc (sizeof (gzochid_oid_holder));
@@ -57,19 +63,10 @@ static gzochid_data_transaction_context *create_transaction_context
   tx_context->context = app_context;
 
   if (timeout != NULL)
-    {
-      tx_context->oids_transaction = 
-	gzochid_storage_transaction_begin_timed (app_context->oids, *timeout);
-      tx_context->names_transaction =
-	gzochid_storage_transaction_begin_timed (app_context->names, *timeout);
-    }
-  else
-    {
-      tx_context->oids_transaction = 
-	gzochid_storage_transaction_begin (app_context->oids);
-      tx_context->names_transaction =
-	gzochid_storage_transaction_begin (app_context->names);
-    }
+    tx_context->transaction = gzochid_storage_transaction_begin_timed 
+      (app_context->storage_context, *timeout);
+  else tx_context->transaction = gzochid_storage_transaction_begin 
+	 (app_context->storage_context);
 
   tx_context->oids_to_references = g_hash_table_new (g_str_hash, g_str_equal);
   tx_context->ptrs_to_references = g_hash_table_new 
@@ -135,8 +132,8 @@ static gboolean flush_reference
 	}
 
       gzochid_storage_transaction_put
-	(context->oids_transaction, oid_str, strlen (oid_str) + 1, 
-	 out->str, out->len);
+	(context->transaction, context->context->oids, oid_str, 
+	 strlen (oid_str) + 1, out->str, out->len);
 
       event = malloc (sizeof (gzochid_application_data_event));
       base_event = (gzochid_application_event *) event;
@@ -186,12 +183,8 @@ static int data_prepare (gpointer data)
   if (flush_failed)
     return FALSE;
 
-  gzochid_storage_transaction_prepare (context->oids_transaction);
-  if (context->oids_transaction->rollback)
-    return FALSE;
-
-  gzochid_storage_transaction_prepare (context->names_transaction);
-  if (context->oids_transaction->rollback)
+  gzochid_storage_transaction_prepare (context->transaction);
+  if (context->transaction->rollback)
     return FALSE;
 
   return TRUE;
@@ -223,8 +216,7 @@ static void data_commit (gpointer data)
   gzochid_data_transaction_context *context = 
     (gzochid_data_transaction_context *) data;
 
-  gzochid_storage_transaction_commit (context->oids_transaction);
-  gzochid_storage_transaction_commit (context->names_transaction);
+  gzochid_storage_transaction_commit (context->transaction);
 
   finalize_references (context);
 
@@ -239,8 +231,7 @@ static void data_rollback (gpointer data)
   gzochid_data_transaction_context *context = 
     (gzochid_data_transaction_context *) data;
 
-  gzochid_storage_transaction_rollback (context->oids_transaction);
-  gzochid_storage_transaction_rollback (context->names_transaction);
+  gzochid_storage_transaction_rollback (context->transaction);
 
   finalize_references (context);
 
@@ -314,35 +305,50 @@ static gzochid_data_oid_block *reserve_oids
 }
 
 static void get_binding 
-(gzochid_data_transaction_context *context, char *name, mpz_t oid)
+(gzochid_data_transaction_context *context, char *name, mpz_t oid, 
+ GError **err)
 {
   char *oid_str = gzochid_storage_transaction_get 
-    (context->names_transaction, name, strlen (name) + 1, NULL);
-  if (context->names_transaction->rollback)
-    gzochid_transaction_mark_for_rollback 
-      (&data_participant, context->names_transaction->should_retry);
+    (context->transaction, context->context->names, name, strlen (name) + 1, 
+     NULL);
 
-  if (oid_str != NULL)
+  if (context->transaction->rollback)
+    {
+      gzochid_transaction_mark_for_rollback 
+	(&data_participant, context->transaction->should_retry);
+      g_set_error
+	(err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_TRANSACTION, 
+	 "Transaction rolled back while retrieving binding.");
+    }
+  else if (oid_str != NULL)
     {
       mpz_set_str (oid, oid_str, 16);
       free (oid_str);
     }
-  else mpz_set_si (oid, -1);
+  else g_set_error
+	(err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_NOT_FOUND, 
+	 "No data found for binding '%s'.", name);
 }
 
 static void set_binding 
-(gzochid_data_transaction_context *context, char *name, mpz_t oid)
+(gzochid_data_transaction_context *context, char *name, mpz_t oid, 
+ GError **err)
 {
   char *oid_str = mpz_get_str (NULL, 16, oid);
   gzochid_storage_transaction_put 
-    (context->names_transaction, name, strlen (name) + 1, oid_str, 
-     strlen (oid_str) + 1);
-
-  if (context->names_transaction->rollback)
-    gzochid_transaction_mark_for_rollback 
-      (&data_participant, context->names_transaction->should_retry);
+    (context->transaction, context->context->names, name, strlen (name) + 1, 
+     oid_str, strlen (oid_str) + 1);
 
   free (oid_str);
+
+  if (context->transaction->rollback)
+    {
+      gzochid_transaction_mark_for_rollback 
+	(&data_participant, context->transaction->should_retry);
+      g_set_error 
+	(err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_TRANSACTION, 
+	 "Transaction rolled back while setting binding.");
+    }
 }
 
 static gzochid_data_managed_reference *create_empty_reference
@@ -451,9 +457,9 @@ static gzochid_data_managed_reference *get_reference_by_ptr
   return reference;
 }
 
-void dereference 
+static void dereference 
 (gzochid_data_transaction_context *context, 
- gzochid_data_managed_reference *reference)
+ gzochid_data_managed_reference *reference, GError **err)
 {
   size_t data_len = 0;
   char *oid_str = NULL; 
@@ -468,14 +474,24 @@ void dereference
   gzochid_debug ("Retrieving data for reference '%s'.", oid_str);
 
   data = gzochid_storage_transaction_get
-    (context->oids_transaction, oid_str, strlen (oid_str) + 1, &data_len);
-  if (context->oids_transaction->rollback)
-    gzochid_transaction_mark_for_rollback 
-      (&data_participant, context->oids_transaction->should_retry);
+    (context->transaction, context->context->oids, oid_str, 
+     strlen (oid_str) + 1, &data_len);
 
-  if (data == NULL)
+  if (context->transaction->rollback)
     {
-      gzochid_info ("No data found for reference '%s'.", oid_str);
+      gzochid_transaction_mark_for_rollback 
+	(&data_participant, context->transaction->should_retry);
+      g_set_error 
+	(err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_TRANSACTION, 
+	 "Transaction rolled back while retrieving reference '%s'.", oid_str);
+      free (oid_str);
+    }
+  else if (data == NULL)
+    {
+      g_set_error 
+	(err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_NOT_FOUND, 
+	 "No data found for reference '%s'.", oid_str);
+
       free (oid_str);
     }
   else 
@@ -503,46 +519,76 @@ void dereference
 
       g_string_free (in, TRUE);
     }
+
+  return;
 }
 
-static void remove_object (gzochid_data_transaction_context *context, mpz_t oid)
+static void remove_object 
+(gzochid_data_transaction_context *context, mpz_t oid, GError **err)
 {
+  int ret = 0;
   char *oid_str = mpz_get_str (NULL, 16, oid);
+
+  /* The transaction is presumed to have already been joined at this point. */
+  
+  gzochid_data_transaction_context *tx_context =
+    gzochid_transaction_context (&data_participant);
+
   gzochid_debug ("Removing reference '%s'.", oid_str);      
-  gzochid_storage_transaction_delete 
-    (context->oids_transaction, oid_str, strlen (oid_str) + 1);
+  ret = gzochid_storage_transaction_delete 
+    (context->transaction, context->context->oids, oid_str, 
+     strlen (oid_str) + 1);
+
+  if (ret == GZOCHID_STORAGE_ENOTFOUND)
+    g_set_error (err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_NOT_FOUND, 
+		 "No data found for reference '%s'.", oid_str);
+  else if (ret == GZOCHID_STORAGE_ETXFAILURE)
+    {
+      gzochid_transaction_mark_for_rollback 
+	(&data_participant, tx_context->transaction->should_retry);
+      g_set_error 
+	(err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_TRANSACTION, 
+	 "Transaction rolled back while removing reference '%s'.", oid_str);
+    }
+
   free (oid_str);
+  return;
 }
 
 static void remove_reference 
 (gzochid_data_transaction_context *context, 
- gzochid_data_managed_reference *reference)
+ gzochid_data_managed_reference *reference, GError **err)
 {
   switch (reference->state)
     {
     case GZOCHID_MANAGED_REFERENCE_STATE_EMPTY:
       
-      remove_object (context, reference->oid);
+      remove_object (context, reference->oid, err);
       reference->state = GZOCHID_MANAGED_REFERENCE_STATE_REMOVED_EMPTY;
+
       break;
 
     case GZOCHID_MANAGED_REFERENCE_STATE_NOT_MODIFIED:
     case GZOCHID_MANAGED_REFERENCE_STATE_MODIFIED:
 
-      remove_object (context, reference->oid);
+      remove_object (context, reference->oid, err);
 
     case GZOCHID_MANAGED_REFERENCE_STATE_NEW:
       reference->state = GZOCHID_MANAGED_REFERENCE_STATE_REMOVED_FETCHED;
     default: break;
     }
+
+  return;
 }
 
-static void join_transaction (gzochid_application_context *context)
+static gzochid_data_transaction_context *join_transaction 
+(gzochid_application_context *context)
 {
+  gzochid_data_transaction_context *tx_context = NULL;
+
   if (!gzochid_transaction_active ()
-      || gzochid_transaction_context (&data_participant) == NULL)
+      || (tx_context = gzochid_transaction_context (&data_participant)) == NULL)
     {
-      gzochid_data_transaction_context *tx_context = NULL;
       if (gzochid_transaction_timed ())
 	{
 	  struct timeval remaining = gzochid_transaction_time_remaining ();
@@ -551,24 +597,25 @@ static void join_transaction (gzochid_application_context *context)
       else tx_context = create_transaction_context (context, NULL); 
       gzochid_transaction_join (&data_participant, tx_context);
     }
+
+  return tx_context;
 }
 
 void *gzochid_data_get_binding
 (gzochid_application_context *context, char *name, 
- gzochid_io_serialization *serialization)
+ gzochid_io_serialization *serialization, GError **err)
 {
   mpz_t oid;
   gzochid_data_managed_reference *reference = NULL;
-  gzochid_data_transaction_context *tx_context = NULL;
-
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&data_participant);
-
+  gzochid_data_transaction_context *tx_context = join_transaction (context);
+  GError *local_err = NULL;
+  
   mpz_init (oid);
-  get_binding (tx_context, name, oid);
+  get_binding (tx_context, name, oid, &local_err);
 
-  if (mpz_cmp_si (oid, -1) == 0)
+  if (local_err != NULL)
     {
+      g_propagate_error (err, local_err);
       mpz_clear (oid);
       return NULL;
     }
@@ -576,105 +623,104 @@ void *gzochid_data_get_binding
   reference = get_reference_by_oid (context, oid, serialization);  
   mpz_clear (oid);
 
-  dereference (tx_context, reference);
+  dereference (tx_context, reference, err);
   return reference->obj;
 }
 
 void gzochid_data_set_binding_to_oid
-(gzochid_application_context *context, char *name, mpz_t oid)
+(gzochid_application_context *context, char *name, mpz_t oid, GError **err)
 {
-  gzochid_data_transaction_context *tx_context = NULL;
- 
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&data_participant);
-  set_binding (tx_context, name, oid);
+  gzochid_data_transaction_context *tx_context = join_transaction (context);
+  set_binding (tx_context, name, oid, err);
 }
 
 char *gzochid_data_next_binding_oid
-(gzochid_application_context *context, char *key, mpz_t oid)
+(gzochid_application_context *context, char *key, mpz_t oid, GError **err)
 {
   char *next_key = NULL;
-  gzochid_data_transaction_context *tx_context = NULL;
- 
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&data_participant);
+  gzochid_data_transaction_context *tx_context = join_transaction (context);
+
   next_key = gzochid_storage_transaction_next_key 
-    (tx_context->names_transaction, key, strlen (key) + 1, NULL);
-  if (tx_context->names_transaction->rollback)
+    (tx_context->transaction, tx_context->context->names, key, 
+     strlen (key) + 1, NULL);
+  if (tx_context->transaction->rollback)
     gzochid_transaction_mark_for_rollback 
-      (&data_participant, tx_context->names_transaction->should_retry);
+      (&data_participant, tx_context->transaction->should_retry);
 
   if (next_key != NULL)
     {
       char *next_value = gzochid_storage_transaction_get 
-	(tx_context->names_transaction, next_key, strlen (next_key) + 1, NULL);
-      if (tx_context->names_transaction->rollback)
+	(tx_context->transaction, tx_context->context->names, next_key, 
+	 strlen (next_key) + 1, NULL);
+      if (tx_context->transaction->rollback)
 	gzochid_transaction_mark_for_rollback 
-	  (&data_participant, tx_context->names_transaction->should_retry);
+	  (&data_participant, tx_context->transaction->should_retry);
 
-      mpz_set_str (oid, next_value, 16);
-      free (next_value);
-      return next_key;
+      if (next_value != NULL)
+	{
+	  mpz_set_str (oid, next_value, 16);
+	  free (next_value);
+	  return next_key;
+	}
+      else return NULL;
     }
   else return NULL;
 }
 
 void gzochid_data_set_binding 
 (gzochid_application_context *context, char *name, 
- gzochid_io_serialization *serialization, void *data)
+ gzochid_io_serialization *serialization, void *data, GError **err)
 {
-  gzochid_data_managed_reference *reference = NULL;
-  gzochid_data_transaction_context *tx_context = NULL;
- 
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&data_participant);
+  gzochid_data_transaction_context *tx_context = join_transaction (context);
+  gzochid_data_managed_reference *reference = 
+    get_reference_by_ptr (context, data, serialization);
 
-  reference = get_reference_by_ptr (context, data, serialization);
-  set_binding (tx_context, name, reference->oid);
+  set_binding (tx_context, name, reference->oid, err);
 }
 
 void gzochid_data_remove_binding
-(gzochid_application_context *context, char *name)
+(gzochid_application_context *context, char *name, GError **err)
 {
-  mpz_t oid;
-  char *oid_str = NULL;
-  gzochid_data_transaction_context *tx_context = NULL;
+  int ret = 0;
+  gzochid_data_transaction_context *tx_context = join_transaction (context);
 
-  mpz_init (oid);
+  ret = gzochid_storage_transaction_delete 
+    (tx_context->transaction, tx_context->context->names, name, 
+     strlen (name) + 1);
 
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&data_participant);
-
-  get_binding (tx_context, name, oid);
-  oid_str = mpz_get_str (NULL, 16, oid);
-  mpz_clear (oid);
-
-  gzochid_storage_transaction_delete 
-    (tx_context->names_transaction, name, strlen (name) + 1);
-  if (tx_context->names_transaction->rollback)
-    gzochid_transaction_mark_for_rollback 
-      (&data_participant, tx_context->names_transaction->should_retry);
-
-  free (oid_str);
+  if (ret == GZOCHID_STORAGE_ENOTFOUND)
+    g_set_error (err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_NOT_FOUND, 
+		 "No data found for binding '%s'.", name);
+  else if (ret == GZOCHID_STORAGE_ETXFAILURE)
+    {
+      gzochid_transaction_mark_for_rollback 
+	(&data_participant, tx_context->transaction->should_retry);
+      g_set_error 
+	(err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_TRANSACTION, 
+	 "Transaction rolled back while removing binding '%s'.", name);
+    }
 }
 
 gboolean gzochid_data_binding_exists
-(gzochid_application_context *context, char *name)
+(gzochid_application_context *context, char *name, GError **err)
 {
   mpz_t oid;
   gboolean ret = TRUE;
-
-  gzochid_data_transaction_context *tx_context = NULL;
+  gzochid_data_transaction_context *tx_context = join_transaction (context);
+  GError *local_err = NULL;
 
   mpz_init (oid);
 
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&data_participant);
+  get_binding (tx_context, name, oid, &local_err);
 
-  get_binding (tx_context, name, oid);
-  
-  if (mpz_cmp_si (oid, -1) == 0)
-    ret = FALSE;
+  if (local_err != NULL)
+    {
+      if (local_err->code != GZOCHID_DATA_ERROR_NOT_FOUND)       
+	g_propagate_error (err, local_err);
+      else g_error_free (local_err);
+
+      ret = FALSE;
+    }
 
   mpz_clear (oid);
   return ret;
@@ -688,80 +734,6 @@ gzochid_data_managed_reference *gzochid_data_create_reference
   return get_reference_by_ptr (context, ptr, serialization);
 }
 
-static void persist_data (gpointer data)
-{
-  gzochid_data_managed_reference_holder *reference_holder = 
-    (gzochid_data_managed_reference_holder *) data;
-
-  reference_holder->reference = gzochid_data_create_reference 
-    (reference_holder->context, 
-     reference_holder->serialization, 
-     reference_holder->data);
-}
-
-static void application_persist_data 
-(gzochid_application_context *context, gzochid_auth_identity *identity, 
- gpointer data)
-{
-  if (gzochid_transaction_active ())
-    persist_data (data);
-  else 
-    {
-      gzochid_data_managed_reference_holder *reference_holder = 
-	(gzochid_data_managed_reference_holder *) data;
-      
-      gzochid_data_managed_reference_holder sub_reference_holder;
-      
-      sub_reference_holder.context = reference_holder->context;
-      sub_reference_holder.data = reference_holder->data;
-      sub_reference_holder.reference = NULL;
-      sub_reference_holder.serialization = reference_holder->serialization;
-
-      gzochid_application_transaction_execute 
-	(context, persist_data, &sub_reference_holder);
-      reference_holder->reference = sub_reference_holder.reference;
-    }
-}
-
-gzochid_data_managed_reference *gzochid_data_create_reference_sync
-(gzochid_application_context *context, gzochid_auth_identity *identity,
- gzochid_io_serialization *serialization, void *ptr)
-{
-  gzochid_data_managed_reference_holder reference_holder;
-
-  reference_holder.context = context;
-  reference_holder.data = ptr;
-  reference_holder.reference = NULL;
-  reference_holder.serialization = serialization;
-
-  if (gzochid_transaction_active ())
-    {
-      gzochid_game_context *game_context = 
-	(gzochid_game_context *) ((gzochid_context *) context)->parent;
-
-      gzochid_application_task application_task;
-      gzochid_task task;
-
-      application_task.worker = application_persist_data;
-      application_task.context = context;
-      application_task.identity = identity;
-      application_task.data = &reference_holder;
-
-      task.worker = gzochid_application_task_thread_worker;
-      task.data = &application_task;
-
-      gzochid_schedule_run_task (game_context->task_queue, &task);
-
-      /* Clone the reference in to the current transaction. */
-      
-      reference_holder.reference = gzochid_data_create_reference_to_oid 
-	(context, serialization, reference_holder.reference->oid);
-    }
-  else gzochid_application_transaction_execute 
-	 (context, persist_data, &reference_holder);
-  return reference_holder.reference;
-}
-
 gzochid_data_managed_reference *gzochid_data_create_reference_to_oid
 (gzochid_application_context *context, 
  gzochid_io_serialization *serialization, mpz_t oid)
@@ -770,44 +742,51 @@ gzochid_data_managed_reference *gzochid_data_create_reference_to_oid
   return get_reference_by_oid (context, oid, serialization);
 }
 
-void gzochid_data_dereference (gzochid_data_managed_reference *reference)
+void gzochid_data_dereference 
+(gzochid_data_managed_reference *reference, GError **err)
 {
-  join_transaction (reference->context);
-  dereference (gzochid_transaction_context (&data_participant), reference);
+  dereference (join_transaction (reference->context), reference, err);
 }
 
-void gzochid_data_remove_object (gzochid_data_managed_reference *reference)
+void gzochid_data_remove_object 
+(gzochid_data_managed_reference *reference, GError **err)
 {
-  join_transaction (reference->context);
-  remove_reference (gzochid_transaction_context (&data_participant), reference);
+  remove_reference (join_transaction (reference->context), reference, err);
 }
 
 void gzochid_data_mark 
 (gzochid_application_context *context, gzochid_io_serialization *serialization,
- void *ptr)
+ void *ptr, GError **err)
 {
+  gzochid_data_transaction_context *tx_context = join_transaction (context);
   gzochid_data_managed_reference *reference = NULL;
-  join_transaction (context);
 
   reference = get_reference_by_ptr (context, ptr, serialization);
 
-  if (reference->state != GZOCHID_MANAGED_REFERENCE_STATE_NEW)
+  if (reference->state != GZOCHID_MANAGED_REFERENCE_STATE_NEW
+      && reference->state != GZOCHID_MANAGED_REFERENCE_STATE_MODIFIED)
     {
       char *data = NULL;
       char *oid_str = mpz_get_str (NULL, 16, reference->oid); 
-      gzochid_data_transaction_context *tx_context = 
-	gzochid_transaction_context (&data_participant);
 
       reference->state = GZOCHID_MANAGED_REFERENCE_STATE_MODIFIED;
       gzochid_debug ("Marking reference '%s' for update.", oid_str);
 
       data = gzochid_storage_transaction_get_for_update
-	(tx_context->oids_transaction, oid_str, strlen (oid_str) + 1, NULL);
-      if (tx_context->oids_transaction->rollback)
-	gzochid_transaction_mark_for_rollback 
-	  (&data_participant, tx_context->oids_transaction->should_retry);
+	(tx_context->transaction, tx_context->context->oids, oid_str, 
+	 strlen (oid_str) + 1, NULL);
+   
+      if (tx_context->transaction->rollback)
+	{
+	  gzochid_transaction_mark_for_rollback 
+	    (&data_participant, tx_context->transaction->should_retry);
+	  g_set_error 
+	    (err, GZOCHID_DATA_ERROR, GZOCHID_DATA_ERROR_TRANSACTION, 
+	     "Transaction rolled back while removing reference '%s'.", 
+	     oid_str);
+	}
+      else free (data);
 
       free (oid_str);
-      free (data);
     }
 }
