@@ -118,6 +118,8 @@ gzochid_protocol_client *gzochid_protocol_client_accept (svz_socket_t *sock)
   mpz_init (client->oid);
   mpz_set_si (client->oid, -1);
 
+  g_mutex_init (&client->sock_mutex);
+  client->write_socket = sock->write_socket;
   client->sock = sock;
   client->connection_description = g_strdup_printf 
     ("%s:%d", "addr", sock->remote_port);
@@ -192,11 +194,12 @@ void gzochid_protocol_client_login_failure (gzochid_protocol_client *client)
 void gzochid_protocol_client_send 
 (gzochid_protocol_client *client, unsigned char *msg, short len)
 {
-  unsigned char len_str[3];
+  unsigned char *buf = malloc (sizeof (unsigned char) * (len + 3));
 
-  gzochi_common_io_write_short (len, len_str, 0);
-  len_str[2] = GZOCHI_COMMON_PROTOCOL_SESSION_MESSAGE;
+  gzochi_common_io_write_short (len, buf, 0);
+  buf[2] = GZOCHI_COMMON_PROTOCOL_SESSION_MESSAGE;
+  memcpy (buf + 3, msg, len);
 
-  svz_sock_write (client->sock, (char *) len_str, 3);
-  svz_sock_write (client->sock, (char *) msg, len);
+  svz_sock_write (client->sock, (char *) buf, len + 3);
+  free (buf);
 }
