@@ -17,6 +17,7 @@
 
 #include <assert.h>
 #include <glib.h>
+#include <gmp.h>
 #include <gzochi-common.h>
 #include <libguile.h>
 #include <stdarg.h>
@@ -71,9 +72,9 @@ static SCM scm_channel_oid;
 
 static SCM scm_make_task_handle;
 static SCM scm_task_handle_oid;
-static SCM scm_run_task;
 
 static SCM scm_r6rs_raise;
+static SCM scm_r6rs_raise_continuable;
 
 static SCM scm_make_object_removed_condition;
 static SCM scm_make_name_exists_condition;
@@ -239,7 +240,7 @@ void gzochid_scheme_application_initialized_worker
       (&scheme_participant,
        is_transaction_retry (scm_variable_ref (exception_var)));
   else gzochid_data_set_binding_to_oid 
-	 (context, "s.initializer", callback_reference->oid);
+	 (context, "s.initializer", callback_reference->oid, NULL);
 }
 
 static gpointer unpack_handler (gpointer data)
@@ -267,13 +268,13 @@ void gzochid_scheme_application_logged_in_worker
 (gzochid_application_context *context, gzochid_auth_identity *identity, 
  gpointer data)
 {
+  GError *err = NULL;
   mpz_t session_oid;
   char *session_oid_str = (char *) data;
   gzochid_client_session *session = NULL;
   gzochid_data_managed_reference *session_reference = NULL;
 
   SCM scm_session = SCM_BOOL_F;
-  gzochid_data_managed_reference *session_scm_reference = NULL;
 
   SCM cb = SCM_BOOL_F;
   gzochid_data_managed_reference *callback_reference = NULL;
@@ -287,14 +288,18 @@ void gzochid_scheme_application_logged_in_worker
     (context, &gzochid_client_session_serialization, session_oid);
   mpz_clear (session_oid);
 
-  gzochid_data_dereference (session_reference);
+  gzochid_data_dereference (session_reference, &err);
+
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+
   session = (gzochid_client_session *) session_reference->obj;
 
   scm_session = gzochid_scheme_create_client_session 
     (session, session_reference->oid);
-  session_scm_reference = 
-    gzochid_data_create_reference
-    (context, &gzochid_scheme_data_serialization, scm_session);
 
   cb = gzochid_scheme_create_callback (context->descriptor->logged_in, NULL);
   callback_reference = gzochid_data_create_reference 
@@ -309,8 +314,7 @@ void gzochid_scheme_application_logged_in_worker
      g_list_append
      (g_list_append
       (g_list_append (NULL, "gzochi"), "private"), "app"),
-     scm_list_2 ((SCM) callback_reference->obj, 
-		 (SCM) session_scm_reference->obj), 
+     scm_list_2 ((SCM) callback_reference->obj, (SCM) scm_session), 
      exception_var);
   
   if (scm_variable_ref (exception_var) != SCM_UNSPECIFIED)
@@ -329,8 +333,20 @@ void gzochid_scheme_application_logged_in_worker
 	(context, identity, unpack_handler, handler);
 
       gzochid_data_mark 
-	(context, &gzochid_client_session_serialization, session);
-      gzochid_client_session_send_login_success (context, session);
+	(context, &gzochid_client_session_serialization, session, &err);
+      if (err == NULL)
+	{
+	  gzochid_scm_location_info *scm_session_reloc = 
+	    gzochid_scm_location_get (context, scm_session);
+	  gzochid_data_managed_reference *reloc_reference = 
+	    gzochid_data_create_reference 
+	    (context, &gzochid_scm_location_aware_serialization, 
+	     scm_session_reloc);
+
+	  mpz_set (session->scm_oid, reloc_reference->oid);
+	  gzochid_client_session_send_login_success (context, session);
+	}
+      else g_error_free (err);
     }
 }
 
@@ -338,6 +354,7 @@ void gzochid_scheme_application_received_message_worker
 (gzochid_application_context *context, gzochid_auth_identity *identity, 
  gpointer ptr)
 {
+  GError *err = NULL;
   gzochid_client_session *session = NULL;
   gzochid_data_managed_reference *session_reference = NULL;
   gzochid_data_managed_reference *callback_reference = NULL;
@@ -360,7 +377,15 @@ void gzochid_scheme_application_received_message_worker
 
   session_reference = gzochid_data_create_reference_to_oid 
     (context, &gzochid_client_session_serialization, session_oid);
-  gzochid_data_dereference (session_reference);
+
+  gzochid_data_dereference (session_reference, &err);
+
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+
   session = (gzochid_client_session *) session_reference->obj;
 
   bv = gzochid_scheme_create_bytevector (message, *message_len_ptr);
@@ -368,7 +393,14 @@ void gzochid_scheme_application_received_message_worker
     gzochid_data_create_reference_to_oid
     (context, &gzochid_scheme_data_serialization, 
      session->handler->received_message->scm_oid);
-  gzochid_data_dereference (callback_reference);
+
+  gzochid_data_dereference (callback_reference, &err);
+
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
       
   gzochid_scheme_invoke
     (context, 
@@ -392,6 +424,7 @@ void gzochid_scheme_application_disconnected_worker
 (gzochid_application_context *context, gzochid_auth_identity *identity, 
  gpointer ptr)
 {
+  GError *err = NULL;
   gzochid_client_session *session = NULL;
   gzochid_data_managed_reference *session_reference = NULL;
   gzochid_data_managed_reference *callback_reference = NULL;
@@ -407,7 +440,15 @@ void gzochid_scheme_application_disconnected_worker
   session_reference = gzochid_data_create_reference_to_oid 
     (context, &gzochid_client_session_serialization, session_oid);
   mpz_clear (session_oid);
-  gzochid_data_dereference (session_reference);
+
+  gzochid_data_dereference (session_reference, &err);
+
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+
   session = (gzochid_client_session *) session_reference->obj;
 
   /* A client may disconnect before the login process has completed or after
@@ -419,7 +460,8 @@ void gzochid_scheme_application_disconnected_worker
 	(context, "Session '%s' disconnected after incomplete login.", ptr);
 
       g_list_free (gpa);
-      gzochid_data_remove_object (session_reference);
+
+      gzochid_data_remove_object (session_reference, NULL);
       return;
     }
 
@@ -427,8 +469,15 @@ void gzochid_scheme_application_disconnected_worker
     gzochid_data_create_reference_to_oid
     (context, &gzochid_scheme_data_serialization, 
      session->handler->disconnected->scm_oid);
-  gzochid_data_dereference (callback_reference);
-      
+
+  gzochid_data_dereference (callback_reference, &err);
+  
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+  
   gzochid_scheme_invoke
     (context, 
      identity,
@@ -443,8 +492,7 @@ void gzochid_scheme_application_disconnected_worker
 	(&scheme_participant, 
 	 is_transaction_retry (scm_variable_ref (exception_var)));
     }
-  else gzochid_data_remove_object (session_reference);
-
+  else gzochid_data_remove_object (session_reference, NULL);
   g_list_free (gpa);
 }
 
@@ -787,6 +835,11 @@ SCM gzochid_scheme_r6rs_raise (SCM cond)
   return scm_call_1 (scm_r6rs_raise, cond);
 }
 
+SCM gzochid_scheme_r6rs_raise_continuable (SCM cond)
+{
+  return scm_call_1 (scm_r6rs_raise_continuable, cond);
+}
+
 SCM gzochid_scheme_make_object_removed_condition ()
 {
   return scm_call_0 (scm_make_object_removed_condition);
@@ -905,6 +958,8 @@ static void bind_scm (char *module, SCM *binding, char *name)
 static void *initialize_bindings (void *ptr)
 {
   bind_scm ("rnrs exceptions", &scm_r6rs_raise, "raise");
+  bind_scm ("rnrs exceptions", &scm_r6rs_raise_continuable, 
+	    "raise-continuable");
 
   bind_scm ("rnrs hashtables", &gzochid_scheme_string_hash, "string-hash");
   bind_scm ("rnrs base", &gzochid_scheme_string_equiv, "equal?");
@@ -960,7 +1015,6 @@ static void *initialize_bindings (void *ptr)
 	    "gzochi:make-task-handle");
   bind_scm ("gzochi private task", &scm_task_handle_oid, 
 	    "gzochi:task-handle-oid");
-  bind_scm ("gzochi private task", &scm_run_task, "gzochi:run-task");
  
   gzochid_api_channel_init ();
   gzochid_api_data_init ();
@@ -1011,12 +1065,3 @@ void gzochid_scheme_initialize_bindings (void)
 
   gzochid_task_register_serialization (&gzochid_scheme_task_serialization);
 }
-
-void gzochid_scheme_append_load_path (char *load_path)
-{
-  SCM scm_load_path = scm_module_variable 
-    (scm_c_resolve_module ("guile"), scm_from_locale_symbol ("%load-path"));
-  scm_variable_set_x 
-    (scm_load_path, scm_cons
-     (scm_from_locale_string (load_path), scm_variable_ref (scm_load_path)));
-}
