@@ -70,6 +70,7 @@ GList *gzochid_scheme_callback_module (SCM);
 SCM gzochid_scheme_callback_data (SCM);
 
 SCM gzochid_scheme_r6rs_raise (SCM);
+SCM gzochid_scheme_r6rs_raise_continuable (SCM);
 SCM gzochid_scheme_make_name_exists_condition (char *);
 SCM gzochid_scheme_make_name_not_bound_condition (char *);
 SCM gzochid_scheme_make_object_removed_condition (void);
@@ -91,6 +92,4 @@ void gzochid_scheme_task_handle_oid (SCM, mpz_t);
 
 void gzochid_scheme_initialize_bindings (void);
 
-void gzochid_scheme_append_load_path (char *);
-
 #endif /* GZOCHID_SCHEME_H */
