@@ -48,7 +48,8 @@
 	  gzochi:make-transaction-retry-condition 
 	  gzochi:transaction-retry-condition? &gzochi:transaction-timeout 
 	  gzochi:make-transaction-timeout-condition
-	  gzochi:transaction-timeout-condition?
+	  gzochi:transaction-timeout-condition? &gzochi:object-removed
+	  gzochi:make-object-removed-condition gzochi:object-removed-condition?
 
           ;; (gzochi data)
 
@@ -70,6 +71,13 @@
 	  gzochi:managed-vector-ref gzochi:managed-vector-set!
           gzochi:managed-vector-length gzochi:managed-vector->list
 
+	  gzochi:make-managed-sequence gzochi:managed-sequence?
+          gzochi:managed-sequence->list gzochi:managed-sequence-add!
+	  gzochi:managed-sequence-contains? gzochi:managed-sequence-delete! 
+	  gzochi:managed-sequence-delete-at! gzochi:managed-sequence-fold-left 
+	  gzochi:managed-sequence-fold-right gzochi:managed-sequence-insert! 
+	  gzochi:managed-sequence-ref gzochi:managed-sequence-size
+
           gzochi:make-managed-hashtable gzochi:managed-hashtable?
           gzochi:managed-hashtable-size gzochi:managed-hashtable-ref
           gzochi:managed-hashtable-set! gzochi:managed-hashtable-delete!
