@@ -18,6 +18,7 @@
 #include <assert.h>
 #include <glib.h>
 #include <gmp.h>
+#include <stddef.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -29,6 +30,7 @@
 #include "io.h"
 #include "log.h"
 #include "protocol.h"
+#include "reloc.h"
 #include "scheme.h"
 #include "session.h"
 #include "task.h"
@@ -329,13 +331,65 @@ gzochid_io_serialization gzochid_client_session_serialization =
     finalize_client_session 
   };
 
+static void remove_session
+(gzochid_application_context *context, const char *oid_str, GError **err)
+{
+  GError *local_err = NULL;
+  GString *binding = g_string_new (SESSION_PREFIX);
+
+  g_string_append (binding, oid_str);
+  gzochid_data_remove_binding (context, binding->str, &local_err);
+  if (local_err == NULL)
+    {
+      mpz_t oid;
+      gzochid_data_managed_reference *session_reference = NULL;
+
+      mpz_init (oid);
+      mpz_set_str (oid, oid_str, 16);
+
+      session_reference = gzochid_data_create_reference_to_oid 
+	(context, &gzochid_client_session_serialization, oid);
+      gzochid_data_dereference (session_reference, &local_err);
+
+      if (local_err == NULL)
+	{
+	  gzochid_client_session *session = 
+	    (gzochid_client_session *) session_reference->obj;
+	  gzochid_data_managed_reference *scm_session_reference =
+	    gzochid_data_create_reference_to_oid 
+	    (context, &gzochid_scm_location_aware_serialization, 
+	     session->scm_oid);
+	  
+	  gzochid_data_remove_object (scm_session_reference, &local_err);
+
+	  if (local_err == NULL)
+	    gzochid_data_remove_object (session_reference, &local_err);
+	  
+	  if (local_err != NULL)
+	    g_propagate_error (err, local_err);	  
+	}
+
+      mpz_clear (oid);
+    }
+  else g_propagate_error (err, local_err);
+   
+  g_string_free (binding, FALSE);
+}
+
+void gzochid_client_session_disconnected_worker
+(gzochid_application_context *context, gzochid_auth_identity *identity,
+ gpointer data)
+{
+  remove_session (context, (const char *) data, NULL);
+}
+
 gzochid_client_session *gzochid_client_session_new 
 (gzochid_auth_identity *identity)
 {
   gzochid_client_session *session = (gzochid_client_session *)
     calloc (1, sizeof (gzochid_client_session));
 
-  session->identity = identity;
+  session->identity = gzochid_auth_identity_clone (identity);
   session->channels = g_sequence_new (NULL);
   mpz_init (session->scm_oid);
 
@@ -349,48 +403,52 @@ void gzochid_client_session_free (gzochid_client_session *session)
   free (session);
 }
 
-static void join_transaction (gzochid_application_context *context)
+static gzochid_client_session_transaction_context *join_transaction 
+(gzochid_application_context *context)
 {
-  if (!gzochid_transaction_active()
-      || gzochid_transaction_context (&session_participant) == NULL)
-    gzochid_transaction_join
-      (&session_participant, create_transaction_context (context));
+  gzochid_client_session_transaction_context *tx_context = NULL;
+
+  if (! gzochid_transaction_active ()
+      || (tx_context = gzochid_transaction_context (&session_participant)) 
+      == NULL)
+    {
+      tx_context = create_transaction_context (context);
+      gzochid_transaction_join (&session_participant, tx_context);
+    }
+
+  return tx_context;
 }
 
 void gzochid_client_session_disconnect 
 (gzochid_application_context *context, gzochid_client_session *session)
 {
+  GError *err = NULL;
   char *oid_str = NULL;
-  GString *binding = g_string_new (SESSION_PREFIX);
-  gzochid_client_session_transaction_context *tx_context = NULL;
-  gzochid_data_managed_reference *reference = NULL;
-  
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&session_participant);
-  reference = gzochid_data_create_reference 
+  gzochid_client_session_transaction_context *tx_context = 
+    join_transaction (context);
+  gzochid_data_managed_reference *reference = gzochid_data_create_reference 
     (context, &gzochid_client_session_serialization, session);
 
   assert (reference->state != GZOCHID_MANAGED_REFERENCE_STATE_NEW);
 
   oid_str = mpz_get_str (NULL, 16, reference->oid);
-  g_string_append (binding, oid_str);
-  gzochid_data_remove_binding (context, binding->str);
-  g_string_free (binding, FALSE);
-  free (oid_str);
 
-  tx_context->operations = g_list_append 
-    (tx_context->operations, create_disconnect_operation (reference->oid));
+  remove_session (context, oid_str, &err);
+
+  if (err == NULL)
+    tx_context->operations = g_list_append 
+      (tx_context->operations, create_disconnect_operation (reference->oid));
+  else g_error_free (err);
+
+  free (oid_str);
 }
 
 void gzochid_client_session_send_login_success
 (gzochid_application_context *context, gzochid_client_session *session)
 {
-  gzochid_client_session_transaction_context *tx_context = NULL;
-  gzochid_data_managed_reference *reference = NULL;
-
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&session_participant);
-  reference = gzochid_data_create_reference 
+  gzochid_client_session_transaction_context *tx_context =
+    join_transaction (context);
+  gzochid_data_managed_reference *reference = gzochid_data_create_reference 
     (context, &gzochid_client_session_serialization, session);
 
   assert (reference->state != GZOCHID_MANAGED_REFERENCE_STATE_NEW);
@@ -403,12 +461,9 @@ void gzochid_client_session_send_login_success
 void gzochid_client_session_send_login_failure
 (gzochid_application_context *context, gzochid_client_session *session)
 {
-  gzochid_client_session_transaction_context *tx_context = NULL;
-  gzochid_data_managed_reference *reference = NULL;
-
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&session_participant);
-  reference = gzochid_data_create_reference 
+  gzochid_client_session_transaction_context *tx_context =
+    join_transaction (context);
+  gzochid_data_managed_reference *reference = gzochid_data_create_reference 
     (context, &gzochid_client_session_serialization, session);
 
   assert (reference->state != GZOCHID_MANAGED_REFERENCE_STATE_NEW);
@@ -422,12 +477,9 @@ void gzochid_client_session_send_message
 (gzochid_application_context *context, gzochid_client_session *session, 
  unsigned char *msg, short len)
 {
-  gzochid_client_session_transaction_context *tx_context = NULL;
-  gzochid_data_managed_reference *reference = NULL;
-
-  join_transaction (context);
-  tx_context = gzochid_transaction_context (&session_participant);
-  reference = gzochid_data_create_reference 
+  gzochid_client_session_transaction_context *tx_context =
+    join_transaction (context);
+  gzochid_data_managed_reference *reference = gzochid_data_create_reference 
     (context, &gzochid_client_session_serialization, session);
 
   assert (reference->state != GZOCHID_MANAGED_REFERENCE_STATE_NEW);
@@ -514,7 +566,8 @@ static void prefix_binding_persistence_task_worker
   free (oid_str);
 
   gzochid_data_set_binding_to_oid 
-    (context, binding->str, persistence_task->holder->oid);
+    (context, binding->str, persistence_task->holder->oid, NULL);
+
   g_string_free (binding, FALSE);
 }
 
@@ -530,7 +583,8 @@ gzochid_task *gzochid_data_prefix_binding_persistence_task_new
     gzochid_application_task_new 
     (context, identity, prefix_binding_persistence_task_worker, task_data);
   gzochid_transactional_application_task_execution *execution = 
-    gzochid_transactional_application_task_execution_new (transactional_task);
+    gzochid_transactional_application_task_execution_new 
+    (transactional_task, NULL);
   gzochid_application_task *application_task = gzochid_application_task_new 
     (context, identity, 
      gzochid_application_resubmitting_transactional_task_worker, execution);
@@ -579,21 +633,39 @@ void gzochid_client_session_persist
 void gzochid_sweep_client_sessions (gzochid_application_context *context)
 {
   mpz_t oid;
+  GError *err = NULL;
   char *next_binding = NULL;
   int prefix_len = strlen (SESSION_PREFIX), num_sessions = 0;
 
   mpz_init (oid);
   gzochid_tx_info (context, "Sweeping old client sessions.");
-  next_binding = gzochid_data_next_binding_oid (context, SESSION_PREFIX, oid);
+  next_binding = gzochid_data_next_binding_oid 
+    (context, SESSION_PREFIX, oid, &err);
+
+  assert (err == NULL);
 
   while (next_binding != NULL 
          && strncmp (SESSION_PREFIX, next_binding, prefix_len) == 0)
     {
+      char *oid_str = mpz_get_str (NULL, 16, oid);
       char *next_next_binding = NULL;
+      
+      remove_session (context, oid_str, &err);
 
-      gzochid_data_remove_binding (context, next_binding);
+      free (oid_str);
+
+      assert (err == NULL || err->code == GZOCHID_DATA_ERROR_NOT_FOUND);
+
+      if (err != NULL)
+	{
+	  g_error_free (err);
+	  err = NULL;
+	}
+      
       next_next_binding = gzochid_data_next_binding_oid 
-	(context, next_binding, oid);
+	(context, next_binding, oid, &err);
+
+      assert (err == NULL);
 
       free (next_binding);
       next_binding = next_next_binding;
