@@ -107,7 +107,8 @@ gzochid_task *gzochid_task_make_transactional_application_task
   gzochid_application_task *transactional_task = 
     gzochid_application_task_new (context, identity, worker, data);
   gzochid_transactional_application_task_execution *execution = 
-    gzochid_transactional_application_task_execution_new (transactional_task);
+    gzochid_transactional_application_task_execution_new 
+    (transactional_task, NULL);
   gzochid_application_task *application_task = 
     gzochid_application_task_new 
     (context, identity, 
@@ -141,7 +142,7 @@ gzochid_durable_application_task_handle *create_durable_task_handle
 
   durable_task_handle->task_data_reference = task_data_reference;
   durable_task_handle->serialization = serialization;
-  durable_task_handle->identity = identity;
+  durable_task_handle->identity = gzochid_auth_identity_clone (identity);
   durable_task_handle->repeats = FALSE;
   durable_task_handle->period = immediate;
   durable_task_handle->target_execution_time = target_execution_time;
@@ -270,22 +271,52 @@ gzochid_task *gzochid_task_immediate_new
 
 static void durable_task_application_worker
 (gzochid_application_context *, gzochid_auth_identity *, gpointer);
+static void durable_task_cleanup_worker
+(gzochid_application_context *, gzochid_auth_identity *, gpointer);
 
-gzochid_task *wrap_durable_task 
-(gzochid_application_context *context, gzochid_auth_identity *identity,
- gzochid_durable_application_task *durable_task,
- gzochid_durable_application_task_handle *durable_task_handle)
+static gzochid_task *wrap_durable_task 
+(gzochid_application_context *context, mpz_t oid, GError **err)
 {
+  GError *local_err = NULL;
   gzochid_game_context *game_context = (gzochid_game_context *) 
     ((gzochid_context *) context)->parent;
+  gzochid_auth_identity *cloned_identity = NULL;
+
+  gzochid_application_task *transactional_task = NULL;
+  gzochid_application_task *cleanup_task = NULL;
+  gzochid_transactional_application_task_execution *execution = NULL;
+
+  gzochid_application_task *application_task = NULL;
+
+  gzochid_durable_application_task *durable_task = 
+    gzochid_durable_application_task_new (oid);
+  gzochid_durable_application_task_handle *durable_task_handle = NULL;
+  gzochid_data_managed_reference *durable_task_handle_reference = 
+    gzochid_data_create_reference_to_oid 
+    (context, &gzochid_durable_application_task_handle_serialization, oid);
+  gzochid_data_dereference (durable_task_handle_reference, &local_err);
+
+  if (local_err != NULL)
+    {
+      g_propagate_error (err, local_err);
+      return NULL;
+    }
   
-  gzochid_application_task *transactional_task = gzochid_application_task_new
-    (context, identity, durable_task_application_worker, durable_task);
-  gzochid_transactional_application_task_execution *execution = 
-    gzochid_transactional_application_task_timed_execution_new 
-    (transactional_task, game_context->tx_timeout);
-  gzochid_application_task *application_task = gzochid_application_task_new
-    (context, identity,
+  durable_task_handle = (gzochid_durable_application_task_handle *)
+    durable_task_handle_reference->obj;
+  cloned_identity = 
+    gzochid_auth_identity_clone (durable_task_handle->identity);
+
+  transactional_task = gzochid_application_task_new
+    (context, cloned_identity, durable_task_application_worker, durable_task);
+  cleanup_task = gzochid_application_task_new
+    (context, cloned_identity, durable_task_cleanup_worker, 
+     mpz_get_str (NULL, 16, oid));
+  execution = gzochid_transactional_application_task_timed_execution_new 
+    (transactional_task, cleanup_task, game_context->tx_timeout);
+
+  application_task = gzochid_application_task_new
+    (context, cloned_identity,
      gzochid_application_resubmitting_transactional_task_worker, execution);
 
   return gzochid_task_new 
@@ -293,23 +324,6 @@ gzochid_task *wrap_durable_task
      durable_task_handle->target_execution_time);
 }
 
-static gzochid_task *rebuild_durable_task
-(gzochid_application_context *context, mpz_t oid)
-{
-  gzochid_durable_application_task *task = NULL;
-  gzochid_durable_application_task_handle *handle = NULL;
-  gzochid_data_managed_reference *handle_reference = 
-    gzochid_data_create_reference_to_oid 
-    (context, &gzochid_durable_application_task_handle_serialization, oid);
-
-  gzochid_data_dereference (handle_reference);
-
-  handle = (gzochid_durable_application_task_handle *) handle_reference->obj;
-  task = gzochid_durable_application_task_new (oid);
-
-  return wrap_durable_task
-    (context, handle->identity, task, handle);
-}
 
 static void commit_scheduled_task (gpointer data, gpointer user_data)
 {
@@ -359,9 +373,35 @@ static gzochid_task_transaction_context *join_transaction
   return tx_context;
 }
 
+static void remove_durable_task 
+(gzochid_application_context *context, mpz_t oid, GError **err)
+{
+  GError *local_err = NULL;
+  char *oid_str = mpz_get_str (NULL, 16, oid);
+  GString *binding = g_string_new (PENDING_TASK_PREFIX);
+  gzochid_data_managed_reference *reference =
+    gzochid_data_create_reference_to_oid
+    (context, &gzochid_durable_application_task_handle_serialization, oid);   
+  
+  g_string_append (binding, oid_str);      
+  gzochid_data_remove_binding (reference->context, binding->str, &local_err);
+  
+  if (local_err == NULL)
+    {
+      gzochid_data_remove_object (reference, &local_err);
+      if (local_err != NULL)
+	g_propagate_error (err, local_err);
+    }
+  else g_propagate_error (err, local_err);
+
+  g_string_free (binding, FALSE);
+  free (oid_str);
+ }
+
 void gzochid_restart_tasks (gzochid_application_context *context)
 {
   mpz_t oid;
+  GError *err = NULL;
   char *next_binding = NULL;
   int prefix_len = strlen (PENDING_TASK_PREFIX);
   gzochid_task_transaction_context *tx_context = NULL;
@@ -370,21 +410,43 @@ void gzochid_restart_tasks (gzochid_application_context *context)
   mpz_init (oid);
   gzochid_tx_info (context, "Resubmitting durable tasks.");
   next_binding = gzochid_data_next_binding_oid 
-    (context, PENDING_TASK_PREFIX, oid);
+    (context, PENDING_TASK_PREFIX, oid, &err);
 
-  join_transaction (context);
-  tx_context = (gzochid_task_transaction_context *) 
-    gzochid_transaction_context (&task_participant);
+  assert (err == NULL);
+
+  tx_context = join_transaction (context);
 
   while (next_binding != NULL 
 	 && strncmp (PENDING_TASK_PREFIX, next_binding, prefix_len) == 0)
     {
-      char *next_next_binding = 
-	gzochid_data_next_binding_oid (context, next_binding, oid);
-      gzochid_task *task = rebuild_durable_task (context, oid);
-      
-      tx_context->scheduled_tasks = g_list_append 
-	(tx_context->scheduled_tasks, task);
+      char *next_next_binding = NULL;
+      gzochid_data_managed_reference *handle_reference = 
+	gzochid_data_create_reference_to_oid 
+	(context, &gzochid_durable_application_task_handle_serialization, oid);
+
+      gzochid_data_dereference (handle_reference, &err);
+
+      if (err != NULL)
+	{
+	  assert (err->code == GZOCHID_DATA_ERROR_NOT_FOUND);
+	  gzochid_tx_warning 
+	    (context, "Task handle not found for resubmitted task.");
+	  g_error_free (err);
+	  err = NULL;
+	}
+      else 
+	{
+	  tx_context->scheduled_tasks = g_list_append 
+	    (tx_context->scheduled_tasks, 
+	     wrap_durable_task (context, oid, &err));
+
+	  assert (err == NULL);
+	}
+
+      next_next_binding = 
+	gzochid_data_next_binding_oid (context, next_binding, oid, &err);
+
+      assert (err == NULL);
 
       free (next_binding);
       next_binding = next_next_binding;
@@ -403,8 +465,9 @@ void gzochid_restart_tasks (gzochid_application_context *context)
 gzochid_durable_application_task_handle *build_durable_task_handle
 (gzochid_application_task *task, 
  gzochid_application_task_serialization *serialization, 
- struct timeval delay, struct timeval *period)
+ struct timeval delay, struct timeval *period, GError **err)
 {
+  GError *local_err = NULL;
   struct timeval target;
 
   gzochid_data_managed_reference *task_data_reference = NULL;
@@ -440,22 +503,33 @@ gzochid_durable_application_task_handle *build_durable_task_handle
 
   binding_name = strncat (binding_name, PENDING_TASK_PREFIX, prefix_len);
   binding_name = strncat (binding_name, oid_str, oid_strlen);
-  gzochid_data_set_binding_to_oid 
-    (task->context, binding_name, handle_reference->oid);
   
   free (oid_str);
-  free (binding_name);
-  
-  return durable_task_handle;
-}
 
+  gzochid_data_set_binding_to_oid 
+    (task->context, binding_name, handle_reference->oid, &local_err);
+
+  if (local_err != NULL)
+    {
+      g_propagate_error (err, local_err);
+      free (binding_name);
+      return NULL;
+    }
+  else
+    {
+      free (binding_name);
+      return durable_task_handle;
+    }
+}
+ 
 static void durable_task_application_worker 
 (gzochid_application_context *context, gzochid_auth_identity *identity, 
  gpointer data)
 {
+  GError *err = NULL;
   gzochid_durable_application_task *task = 
     (gzochid_durable_application_task *) data;
-
+  gzochid_task_transaction_context *tx_context = NULL;
   gzochid_data_managed_reference *handle_reference =
     gzochid_data_create_reference_to_oid
     (context, &gzochid_durable_application_task_handle_serialization, 
@@ -463,19 +537,53 @@ static void durable_task_application_worker
   gzochid_durable_application_task_handle *handle = NULL;
   gzochid_application_task *inner_task = NULL;
 
-  gzochid_data_dereference (handle_reference);
+  gzochid_data_dereference (handle_reference, &err);
 
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+      
   handle = (gzochid_durable_application_task_handle *) handle_reference->obj;
-  gzochid_data_dereference (handle->task_data_reference);
+
+  gzochid_data_dereference (handle->task_data_reference, &err);
+
+  if (err != NULL)
+    {
+      g_error_free (err);
+      return;
+    }
+
   inner_task = gzochid_application_task_new 
     (context, identity, handle->task_worker, handle->task_data_reference->obj);
-
   inner_task->worker (context, identity, inner_task->data);
+  free (inner_task);
 
-  if (handle->repeats)
+  tx_context = join_transaction (context);
+  if (gzochid_transaction_rollback_only ())
+
+    /* The body of the task worker may have done something to cause the
+       transaction to be rolled back. If that's the case, we should not bother
+       to attempt any logic that involves the data manager. */
+
+    return;
+
+  /* The repeats status of the task may have changed while executing the 
+     worker, so check again to see if it needs to be removed. */
+
+  if (! handle->repeats)
+    {
+      remove_durable_task (context, task->handle_oid, &err);
+
+      if (err != NULL)
+	{
+	  g_error_free (err);
+	  return;
+	}
+    }
+  else
     {
-      gzochid_task_transaction_context *tx_context = NULL;
-      gzochid_task *wrapped_task = NULL;
       struct timeval now;
 
       gettimeofday (&now, NULL);
@@ -485,30 +593,38 @@ static void durable_task_application_worker
 
       gzochid_data_mark 
 	(context, &gzochid_durable_application_task_handle_serialization, 
-	 handle);
-      wrapped_task = wrap_durable_task (context, identity, task, handle);
+	 handle, &err);
 
-      join_transaction (context);
-      tx_context = (gzochid_task_transaction_context *) 
-	gzochid_transaction_context (&task_participant);
-      
-      tx_context->scheduled_tasks = g_list_append 
-	(tx_context->scheduled_tasks, wrapped_task);
-    }
-  else
-    { 
-      char *oid_str = mpz_get_str (NULL, 16, task->handle_oid);
-      GString *binding = g_string_new (PENDING_TASK_PREFIX);
-      
-      g_string_append (binding, oid_str);
-      
-      gzochid_data_remove_binding (context, binding->str);
-      
-      g_string_free (binding, FALSE);
-      free (oid_str);
+      if (err == NULL)
+	{
+	  gzochid_task *new_task = wrap_durable_task 
+	    (context, task->handle_oid, NULL);
+
+	  if (new_task != NULL)
+	    tx_context->scheduled_tasks = g_list_append 
+	      (tx_context->scheduled_tasks, new_task);
+	}
+	  
+      else g_error_free (err);
     }
 }
 
+static void durable_task_cleanup_worker 
+(gzochid_application_context *context, gzochid_auth_identity *identity, 
+ gpointer data)
+{
+  char *oid_str = (char *) data;
+  mpz_t oid;
+
+  mpz_init (oid);
+  mpz_set_str (oid, oid_str, 16);
+  
+  remove_durable_task (context, oid, NULL);
+
+  mpz_clear (oid);
+  free (oid_str);
+}
+
 void gzochid_task_free (gzochid_task *task)
 {
   free (task);
@@ -527,11 +643,7 @@ void gzochid_schedule_durable_task
 static void schedule_durable_task 
 (gzochid_application_context *context, gzochid_task *task)
 {
-  gzochid_task_transaction_context *tx_context = NULL;
-  
-  join_transaction (context);
-  tx_context = (gzochid_task_transaction_context *) 
-    gzochid_transaction_context (&task_participant);
+  gzochid_task_transaction_context *tx_context = join_transaction (context);
 
   tx_context->scheduled_tasks = g_list_append 
     (tx_context->scheduled_tasks, task);
@@ -543,16 +655,21 @@ void gzochid_schedule_delayed_durable_task
  gzochid_application_task_serialization *serialization, 
  struct timeval delay)
 {
+  GError *err = NULL;      
   gzochid_durable_application_task_handle *handle = 
-    build_durable_task_handle (task, serialization, delay, NULL);
-  gzochid_data_managed_reference *handle_reference = 
-    gzochid_data_create_reference 
-    (context, &gzochid_durable_application_task_handle_serialization, handle);
-  gzochid_task *wrapped_task = wrap_durable_task
-    (context, identity,
-     gzochid_durable_application_task_new (handle_reference->oid), handle);
+    build_durable_task_handle (task, serialization, delay, NULL, &err);
 
-  schedule_durable_task (context, wrapped_task);     
+  if (err == NULL)
+    {
+      gzochid_data_managed_reference *handle_reference = 
+	gzochid_data_create_reference 
+	(context, &gzochid_durable_application_task_handle_serialization, 
+	 handle);
+      
+      schedule_durable_task 
+	(context, wrap_durable_task (context, handle_reference->oid, NULL));
+    }
+  else g_error_free (err);
 }
 
 gzochid_periodic_task_handle *gzochid_schedule_periodic_durable_task 
@@ -561,16 +678,21 @@ gzochid_periodic_task_handle *gzochid_schedule_periodic_durable_task
  gzochid_application_task_serialization *serialization, 
  struct timeval delay, struct timeval period)
 {
+  GError *err = NULL;
   gzochid_durable_application_task_handle *handle = 
-    build_durable_task_handle (task, serialization, delay, &period);
-  gzochid_data_managed_reference *handle_reference = 
-    gzochid_data_create_reference 
-    (context, &gzochid_durable_application_task_handle_serialization, handle);
-  gzochid_task *wrapped_task = wrap_durable_task
-    (context, identity,
-     gzochid_durable_application_task_new (handle_reference->oid), handle);
+    build_durable_task_handle (task, serialization, delay, &period, &err);
 
-  schedule_durable_task (context, wrapped_task);
+  if (err == NULL)
+    {
+      gzochid_data_managed_reference *handle_reference = 
+	gzochid_data_create_reference 
+	(context, &gzochid_durable_application_task_handle_serialization, 
+	 handle);
+      
+      schedule_durable_task 
+	(context, wrap_durable_task (context, handle_reference->oid, NULL));
+    }
+  else g_error_free (err);
 
   return handle;
 }
@@ -579,6 +701,8 @@ void gzochid_cancel_periodic_task
 (gzochid_application_context *context, gzochid_periodic_task_handle *handle)
 {
   handle->repeats = FALSE;
+
   gzochid_data_mark 
-    (context, &gzochid_durable_application_task_handle_serialization, handle);
+    (context, &gzochid_durable_application_task_handle_serialization, 
+     handle, NULL);
 }
