@@ -33,9 +33,11 @@ gzochid_application_context *gzochid_application_context_new ()
   context->free_oids_lock = g_mutex_new ();
   context->client_mapping_lock = g_mutex_new ();
 
-  context->meta = gzochid_storage_open ("/dev/null");
-  context->oids = gzochid_storage_open ("/dev/null");
-  context->names = gzochid_storage_open ("/dev/null");
+  context->storage_context = gzochid_storage_initialize ("/dev/null?");
+  context->meta = gzochid_storage_open (context->storage_context, "/dev/null");
+  context->oids = gzochid_storage_open (context->storage_context, "/dev/null");
+  context->names = gzochid_storage_open 
+    (context->storage_context, "/dev/null");
 
   context->event_source = gzochid_application_event_source_new ();
 
@@ -54,12 +56,14 @@ void gzochid_application_task_thread_worker (gpointer data, gpointer user_data)
 }
 
 static gzochid_transactional_application_task_execution *execution_new
-(gzochid_application_task *task, struct timeval *timeout)
+(gzochid_application_task *task, gzochid_application_task *cleanup_task, 
+ struct timeval *timeout)
 {
   gzochid_transactional_application_task_execution *execution = 
     malloc (sizeof (gzochid_transactional_application_task_execution));
 
   execution->task = task;
+  execution->cleanup_task = cleanup_task;
 
   if (timeout != NULL)
     {
@@ -77,16 +81,17 @@ static gzochid_transactional_application_task_execution *execution_new
 
 gzochid_transactional_application_task_execution *
 gzochid_transactional_application_task_execution_new 
-(gzochid_application_task *task)
+(gzochid_application_task *task, gzochid_application_task *cleanup_task)
 {
-  return execution_new (task, NULL);
+  return execution_new (task, cleanup_task, NULL);
 }
 
 gzochid_transactional_application_task_execution *
 gzochid_transactional_application_task_timed_execution_new 
-(gzochid_application_task *task, struct timeval timeout)
+(gzochid_application_task *task, gzochid_application_task *cleanup_task, 
+ struct timeval timeout)
 {
-  return execution_new (task, &timeout);
+  return execution_new (task, cleanup_task, &timeout);
 }
 
 void gzochid_serialize_application_task 
