@@ -412,6 +412,16 @@ gboolean gzochid_transaction_rollback_only ()
   return rollback_only (transaction);
 }
 
+gboolean gzochid_transaction_retryable ()
+{
+  gzochid_transaction *transaction = (gzochid_transaction *) 
+    g_static_private_get (&thread_transaction_key);
+
+  assert (transaction != NULL);
+
+  return retryable (transaction);
+}
+
 gboolean gzochid_transaction_timed_out ()
 {
   gzochid_transaction *transaction = (gzochid_transaction *) 
